/*******************************************************************************
* KDE plasma applet to display network interface status.
*
* Copyright (C) 2012 by Eugene Grigorjev <osmiy.lv@gmail.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*******************************************************************************/

#include "network_status.h"

#include <errno.h>
#include <net/ethernet.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include <QPainter>
#include <QSizeF>
#include <KStandardDirs>
#include <QSvgRenderer>
#include <KLocale>
#include <QLineEdit>
#include <QDir>
#include <QtNetwork>

#include <plasma/svg.h>
#include <plasma/theme.h>

// count of wireless signal steps displayed by images
#define NS_WIRELESS_STEPS 4.0

network_status::network_status(QObject *parent, const QVariantList &args)
        : Plasma::Applet(parent, args),
        m_dialog(0),
        m_timer(0),
        m_isDataChanged(true)
{
    setBackgroundHints(DefaultBackground);
    setHasConfigurationInterface(true);
    setAspectRatioMode(Plasma::Square);

    resize(200, 200);
}

network_status::~network_status()
{
    m_timer->stop();
    delete m_timer;
    delete m_dialog;
}

void network_status::init()
{
    setIfName(config().readEntry("interface", "eth0"));

    m_timer = new QTimer(this);
    m_timer->setInterval(config().readEntry("interval", 2000));

    connect(m_timer, SIGNAL(timeout()), this, SLOT(updateData()));
    m_timer->start();
}

void network_status::showConfigurationInterface()
{
    if (m_dialog == 0)
    {
        m_dialog = new KDialog();
        m_dialog->setCaption(i18n("Configure Network Status"));
        ui.setupUi(m_dialog->mainWidget());
        m_dialog->mainWidget()->layout()->setMargin(0);
        m_dialog->setButtons(KDialog::Ok | KDialog::Cancel | KDialog::Apply);

        connect(m_dialog, SIGNAL(applyClicked()), this, SLOT(configAccepted()));
        connect(m_dialog, SIGNAL(okClicked()), this, SLOT(configAccepted()));
    }

    ui.interfaceName->setText(m_ifName);
    ui.updateInterval->setValue(m_timer->interval());

    m_dialog->show();
}

void network_status::configAccepted()
{
    if (0 != m_ifName.compare(ui.interfaceName->text()))
    {
        setIfName(ui.interfaceName->text());
    }

    m_timer->setInterval(ui.updateInterval->value());


    config().writeEntry("interface", ui.interfaceName->text());
    config().writeEntry("interval", ui.updateInterval->value());

    emit configNeedsSaving();
}

void network_status::setIfName(QString ifName)
{
    if (0 == m_ifName.compare(ifName))
        return;

    m_ifName = ifName;
    m_isDataChanged = true;

    /* only for interface name changing do force data update */
    updateData();
}

void network_status::updateData()
{
    QNetworkInterface iface = QNetworkInterface::interfaceFromName(m_ifName);

    setIsValid(iface.isValid());

    /* TODO: find more correct way, maybe using wireless extension? */
    setIsWireless(QDir(QString("/sys/class/net/%1/wireless").arg(m_ifName)).exists());

    setIsActive(
            iface.isValid() &&
            iface.flags().testFlag(QNetworkInterface::IsUp) &&
            iface.flags().testFlag(QNetworkInterface::IsRunning)
        );

    /* NOTE: have to be called after setIsWireless and setIsActive */
    setLinkQuality(getLinkQuality());

    if (m_isDataChanged)
    {
        update();
    }
}

void network_status::setIsValid(bool isValid)
{
    if (m_isValid == isValid)
        return;

    m_isValid = isValid;
    m_isDataChanged = true;
}


void network_status::setIsWireless(bool isWireless)
{
    if (m_isWireless == isWireless)
        return;

    m_isWireless = isWireless;
    m_isDataChanged = true;
}

void network_status::setIsActive(bool isActive)
{
    if (m_isActive == isActive)
        return;

    m_isActive = isActive;
    m_isDataChanged = true;
}

void network_status::setLinkQuality(int linkQuality)
{
    if (m_linkQuality == linkQuality)
        return;

    m_linkQuality = linkQuality;
    m_isDataChanged = true;
}


int network_status::getLinkQuality()
{
    struct iwreq rq;
    struct iw_range range;
    struct iw_statistics iwstats;

    if ((false == m_isWireless) || (false == m_isActive))
    {
        return 0;
    }

    memset(&rq, 0, sizeof(rq));
    strncpy(rq.ifr_name, m_ifName.toLatin1().constData(), IFNAMSIZ);

    int fd = socket( AF_INET, SOCK_DGRAM, 0 );
    if ( fd < 0 ) {
        qDebug() << "Cannot open signal strength socket" << strerror(errno);
        return 0;
    }

    rq.u.data.flags = 0;
    rq.u.data.pointer = (caddr_t) (&iwstats);
    rq.u.data.length = sizeof( struct iw_statistics );
    memset(&iwstats, 0, sizeof(iwstats));

    if (ioctl(fd, SIOCGIWSTATS, &rq)) {
        qDebug() << "Cannot obtain wireless statistics" << strerror(errno);
        ::close( fd );
        return 0;
    }

    rq.u.data.flags = 0;
    rq.u.data.pointer = (caddr_t) (&range);
    rq.u.data.length = sizeof(range);
    memset(&range, 0, sizeof(range));

    if (ioctl(fd, SIOCGIWRANGE, &rq)) {
        qDebug() << "Cannot obtain wireless range" << strerror(errno);
        ::close( fd );
        return 0;
    }

    ::close( fd );

    if (iwstats.qual.updated & IW_QUAL_QUAL_INVALID)
    {
        qDebug() << "Invalid quality received";
        return 0;
    }

    int linkQuality = (int) (
                        round(
                            ((float)iwstats.qual.qual) * NS_WIRELESS_STEPS /
                            ((float)range.max_qual.qual)
                        ) *
                        (100.0 / NS_WIRELESS_STEPS)
                    );

    qDebug() << QString("Link Quality=%1/%2 (%3)").arg(iwstats.qual.qual).arg(range.max_qual.qual).arg(linkQuality);

    return linkQuality;
}

void network_status::paintInterface(QPainter *p, const QStyleOptionGraphicsItem *option, const QRect &contentsRect)
{
    Q_UNUSED(option);

    p->setRenderHint(QPainter::SmoothPixmapTransform);
    p->setRenderHint(QPainter::Antialiasing);

    QString image("error.svgz");

    if (m_isValid)
    {
        image = QString("network-%1-%2.svgz").arg(m_isWireless ? "wireless" : "wired").
                arg(
                    (false == m_isActive) ? "disconnected" :
                    ((false == m_isWireless) ? "connected" :
                     QString::number(m_linkQuality))
                );
    }

    image.prepend(KStandardDirs::locate("data","plasma/plasmoids/" + this->pluginName() + "/images/"));

    QSvgRenderer(image).render(p,  QRectF(contentsRect.adjusted(2, 2,  -2,  -6)));

    m_isDataChanged = false;
}

#include "network_status.moc"

/* ex: set tabstop=4 shiftwidth=4 expandtab: */
